@@ -10,11 +10,11 @@
  * relations with finite memory space usage.  To do that, we set upper bounds
  * on the number of tuples and pages we will keep track of at once.
  *
- * We are willing to use at most VacuumMem memory space to keep track of
- * dead tuples.  We initially allocate an array of TIDs of that size.
- * If the array threatens to overflow, we suspend the heap scan phase
- * and perform a pass of index cleanup and page compaction, then resume
- * the heap scan with an empty TID array.
+ * We are willing to use at most maintenance_work_mem memory space to keep
+ * track of dead tuples.  We initially allocate an array of TIDs of that size.
+ * If the array threatens to overflow, we suspend the heap scan phase and
+ * perform a pass of index cleanup and page compaction, then resume the heap
+ * scan with an empty TID array.
  *
  * We can limit the storage for page free space to MaxFSMPages entries,
  * since that's the most the free space map will be willing to remember
@@ -26,17 +26,19 @@
  * are passed to the free space map at conclusion of the scan.
  *
  *
- * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql-server/src/backend/commands/vacuumlazy.c,v 1.32 2003/09/25 06:57:59 petere Exp $
+ *	  $PostgreSQL: pgsql/src/backend/commands/vacuumlazy.c,v 1.50 2004/12/31 21:59:42 pgsql Exp $
  *
  *-------------------------------------------------------------------------
  */
 #include "postgres.h"
 
+#include <math.h>
+
 #include "access/genam.h"
 #include "access/heapam.h"
 #include "access/xlog.h"
@@ -67,6 +69,8 @@ typedef struct LVRelStats
 	/* Overall statistics about rel */
 	BlockNumber rel_pages;
 	double		rel_tuples;
+	BlockNumber	pages_removed;
+	double		tuples_deleted;
 	BlockNumber nonempty_pages; /* actually, last nonempty page + 1 */
 	Size		threshold;		/* minimum interesting free space */
 	/* List of TIDs of tuples we intend to delete */
@@ -94,12 +98,19 @@ static void lazy_scan_heap(Relation onerel, LVRelStats *vacrelstats,
 			   Relation *Irel, int nindexes);
 static void lazy_vacuum_heap(Relation onerel, LVRelStats *vacrelstats);
 static void lazy_scan_index(Relation indrel, LVRelStats *vacrelstats);
-static void lazy_vacuum_index(Relation indrel, LVRelStats *vacrelstats);
+static void lazy_vacuum_index(Relation indrel,
+							  double *index_tups_vacuumed,
+							  BlockNumber *index_pages_removed,
+							  LVRelStats *vacrelstats);
 static int lazy_vacuum_page(Relation onerel, BlockNumber blkno, Buffer buffer,
 				 int tupindex, LVRelStats *vacrelstats);
 static void lazy_truncate_heap(Relation onerel, LVRelStats *vacrelstats);
 static BlockNumber count_nondeletable_pages(Relation onerel,
 						 LVRelStats *vacrelstats);
+static void lazy_update_relstats(Relation rel, BlockNumber num_pages,
+								 BlockNumber pages_removed,
+								 double num_tuples, double tuples_removed,
+								 bool hasindex);
 static void lazy_space_alloc(LVRelStats *vacrelstats, BlockNumber relblocks);
 static void lazy_record_dead_tuple(LVRelStats *vacrelstats,
 					   ItemPointer itemptr);
@@ -145,14 +156,14 @@ lazy_vacuum_rel(Relation onerel, VacuumStmt *vacstmt)
 	vacrelstats->threshold = GetAvgFSMRequestSize(&onerel->rd_node);
 
 	/* Open all indexes of the relation */
-	vac_open_indexes(onerel, &nindexes, &Irel);
+	vac_open_indexes(onerel, ShareUpdateExclusiveLock, &nindexes, &Irel);
 	hasindex = (nindexes > 0);
 
 	/* Do the vacuuming */
 	lazy_scan_heap(onerel, vacrelstats, Irel, nindexes);
 
 	/* Done with indexes */
-	vac_close_indexes(nindexes, Irel);
+	vac_close_indexes(nindexes, Irel, NoLock);
 
 	/*
 	 * Optionally truncate the relation.
@@ -169,8 +180,10 @@ lazy_vacuum_rel(Relation onerel, VacuumStmt *vacstmt)
 	lazy_update_fsm(onerel, vacrelstats);
 
 	/* Update statistics in pg_class */
-	vac_update_relstats(RelationGetRelid(onerel), vacrelstats->rel_pages,
-						vacrelstats->rel_tuples, hasindex);
+	lazy_update_relstats(onerel, vacrelstats->rel_pages,
+						 vacrelstats->pages_removed,
+						 vacrelstats->rel_tuples, vacrelstats->tuples_deleted,
+						 hasindex);
 }
 
 
@@ -195,6 +208,9 @@ lazy_scan_heap(Relation onerel, LVRelStats *vacrelstats,
 				tups_vacuumed,
 				nkeep,
 				nunused;
+	double	   *index_tups_vacuumed;
+	BlockNumber *index_pages_removed;
+	bool		did_vacuum_index = false;
 	int			i;
 	VacRUsage	ru0;
 
@@ -209,6 +225,16 @@ lazy_scan_heap(Relation onerel, LVRelStats *vacrelstats,
 	empty_pages = 0;
 	num_tuples = tups_vacuumed = nkeep = nunused = 0;
 
+	/*
+	 * Because index vacuuming is done in multiple passes, we have to keep
+	 * track of the total number of rows and pages removed from each index.
+	 * index_tups_vacuumed[i] is the number removed so far from the i'th
+	 * index.  (For partial indexes this could well be different from
+	 * tups_vacuumed.)  Likewise for index_pages_removed[i].
+	 */
+	index_tups_vacuumed = (double *) palloc0(nindexes * sizeof(double));
+	index_pages_removed = (BlockNumber *) palloc0(nindexes * sizeof(BlockNumber));
+
 	nblocks = RelationGetNumberOfBlocks(onerel);
 	vacrelstats->rel_pages = nblocks;
 	vacrelstats->nonempty_pages = 0;
@@ -226,7 +252,7 @@ lazy_scan_heap(Relation onerel, LVRelStats *vacrelstats,
 					hastup;
 		int			prev_dead_count;
 
-		CHECK_FOR_INTERRUPTS();
+		vacuum_delay_point();
 
 		/*
 		 * If we are close to overrunning the available space for
@@ -238,7 +264,11 @@ lazy_scan_heap(Relation onerel, LVRelStats *vacrelstats,
 		{
 			/* Remove index entries */
 			for (i = 0; i < nindexes; i++)
-				lazy_vacuum_index(Irel[i], vacrelstats);
+				lazy_vacuum_index(Irel[i],
+								  &index_tups_vacuumed[i],
+								  &index_pages_removed[i],
+								  vacrelstats);
+			did_vacuum_index = true;
 			/* Remove tuples from heap */
 			lazy_vacuum_heap(onerel, vacrelstats);
 			/* Forget the now-vacuumed tuples, and press on */
@@ -291,7 +321,6 @@ lazy_scan_heap(Relation onerel, LVRelStats *vacrelstats,
 			 offnum = OffsetNumberNext(offnum))
 		{
 			ItemId		itemid;
-			uint16		sv_infomask;
 
 			itemid = PageGetItemId(page, offnum);
 
@@ -307,9 +336,8 @@ lazy_scan_heap(Relation onerel, LVRelStats *vacrelstats,
 			ItemPointerSet(&(tuple.t_self), blkno, offnum);
 
 			tupgone = false;
-			sv_infomask = tuple.t_data->t_infomask;
 
-			switch (HeapTupleSatisfiesVacuum(tuple.t_data, OldestXmin))
+			switch (HeapTupleSatisfiesVacuum(tuple.t_data, OldestXmin, buf))
 			{
 				case HEAPTUPLE_DEAD:
 					tupgone = true;		/* we can delete the tuple */
@@ -336,6 +364,14 @@ lazy_scan_heap(Relation onerel, LVRelStats *vacrelstats,
 						Assert(tuple.t_data->t_infomask & HEAP_XMIN_COMMITTED);
 						pgchanged = true;
 					}
+
+					/*
+					 * Other checks...
+					 */
+					if (onerel->rd_rel->relhasoids &&
+						!OidIsValid(HeapTupleGetOid(&tuple)))
+						elog(WARNING, "relation \"%s\" TID %u/%u: OID is invalid",
+							 relname, blkno, offnum);
 					break;
 				case HEAPTUPLE_RECENTLY_DEAD:
 
@@ -356,18 +392,6 @@ lazy_scan_heap(Relation onerel, LVRelStats *vacrelstats,
 					break;
 			}
 
-			/* check for hint-bit update by HeapTupleSatisfiesVacuum */
-			if (sv_infomask != tuple.t_data->t_infomask)
-				pgchanged = true;
-
-			/*
-			 * Other checks...
-			 */
-			if (onerel->rd_rel->relhasoids &&
-				!OidIsValid(HeapTupleGetOid(&tuple)))
-				elog(WARNING, "relation \"%s\" TID %u/%u: OID is invalid",
-					 relname, blkno, offnum);
-
 			if (tupgone)
 			{
 				lazy_record_dead_tuple(vacrelstats, &(tuple.t_self));
@@ -399,13 +423,14 @@ lazy_scan_heap(Relation onerel, LVRelStats *vacrelstats,
 		LockBuffer(buf, BUFFER_LOCK_UNLOCK);
 
 		if (pgchanged)
-			SetBufferCommitInfoNeedsSave(buf);
-
-		ReleaseBuffer(buf);
+			WriteBuffer(buf);
+		else
+			ReleaseBuffer(buf);
 	}
 
 	/* save stats for use later */
 	vacrelstats->rel_tuples = num_tuples;
+	vacrelstats->tuples_deleted = tups_vacuumed;
 
 	/* If any tuples need to be deleted, perform final vacuum cycle */
 	/* XXX put a threshold on min number of tuples here? */
@@ -413,11 +438,14 @@ lazy_scan_heap(Relation onerel, LVRelStats *vacrelstats,
 	{
 		/* Remove index entries */
 		for (i = 0; i < nindexes; i++)
-			lazy_vacuum_index(Irel[i], vacrelstats);
+			lazy_vacuum_index(Irel[i],
+							  &index_tups_vacuumed[i],
+							  &index_pages_removed[i],
+							  vacrelstats);
 		/* Remove tuples from heap */
 		lazy_vacuum_heap(onerel, vacrelstats);
 	}
-	else
+	else if (!did_vacuum_index)
 	{
 		/* Must do post-vacuum cleanup and statistics update anyway */
 		for (i = 0; i < nindexes; i++)
@@ -467,7 +495,7 @@ lazy_vacuum_heap(Relation onerel, LVRelStats *vacrelstats)
 		Buffer		buf;
 		Page		page;
 
-		CHECK_FOR_INTERRUPTS();
+		vacuum_delay_point();
 
 		tblk = ItemPointerGetBlockNumber(&vacrelstats->dead_tuples[tupindex]);
 		buf = ReadBuffer(onerel, tblk);
@@ -532,7 +560,7 @@ lazy_vacuum_page(Relation onerel, BlockNumber blkno, Buffer buffer,
 
 		recptr = log_heap_clean(onerel, buffer, unused, uncnt);
 		PageSetLSN(page, recptr);
-		PageSetSUI(page, ThisStartUpID);
+		PageSetTLI(page, ThisTimeLineID);
 	}
 	else
 	{
@@ -560,9 +588,12 @@ lazy_scan_index(Relation indrel, LVRelStats *vacrelstats)
 	vac_init_rusage(&ru0);
 
 	/*
-	 * If index is unsafe for concurrent access, must lock it.
+	 * Acquire appropriate type of lock on index: must be exclusive if
+	 * index AM isn't concurrent-safe.
 	 */
-	if (!indrel->rd_am->amconcurrent)
+	if (indrel->rd_am->amconcurrent)
+		LockRelation(indrel, RowExclusiveLock);
+	else
 		LockRelation(indrel, AccessExclusiveLock);
 
 	/*
@@ -582,26 +613,28 @@ lazy_scan_index(Relation indrel, LVRelStats *vacrelstats)
 	/*
 	 * Release lock acquired above.
 	 */
-	if (!indrel->rd_am->amconcurrent)
+	if (indrel->rd_am->amconcurrent)
+		UnlockRelation(indrel, RowExclusiveLock);
+	else
 		UnlockRelation(indrel, AccessExclusiveLock);
 
 	if (!stats)
 		return;
 
 	/* now update statistics in pg_class */
-	vac_update_relstats(RelationGetRelid(indrel),
-						stats->num_pages, stats->num_index_tuples,
-						false);
+	lazy_update_relstats(indrel, stats->num_pages, stats->pages_removed,
+						 stats->num_index_tuples, stats->tuples_removed,
+						 false);
 
 	ereport(elevel,
-			(errmsg("index \"%s\" now contains %.0f row versions in %u pages",
-					RelationGetRelationName(indrel),
-					stats->num_index_tuples,
-					stats->num_pages),
-			 errdetail("%u index pages have been deleted, %u are currently reusable.\n"
-					   "%s",
-					   stats->pages_deleted, stats->pages_free,
-					   vac_show_rusage(&ru0))));
+	   (errmsg("index \"%s\" now contains %.0f row versions in %u pages",
+			   RelationGetRelationName(indrel),
+			   stats->num_index_tuples,
+			   stats->num_pages),
+		errdetail("%u index pages have been deleted, %u are currently reusable.\n"
+				  "%s",
+				  stats->pages_deleted, stats->pages_free,
+				  vac_show_rusage(&ru0))));
 
 	pfree(stats);
 }
@@ -612,11 +645,17 @@ lazy_scan_index(Relation indrel, LVRelStats *vacrelstats)
  *		Delete all the index entries pointing to tuples listed in
  *		vacrelstats->dead_tuples.
  *
+ *		Increment *index_tups_vacuumed by the number of index entries
+ *		removed, and *index_pages_removed by the number of pages removed.
+ *
  *		Finally, we arrange to update the index relation's statistics in
  *		pg_class.
  */
 static void
-lazy_vacuum_index(Relation indrel, LVRelStats *vacrelstats)
+lazy_vacuum_index(Relation indrel,
+				  double *index_tups_vacuumed,
+				  BlockNumber *index_pages_removed,
+				  LVRelStats *vacrelstats)
 {
 	IndexBulkDeleteResult *stats;
 	IndexVacuumCleanupInfo vcinfo;
@@ -625,9 +664,12 @@ lazy_vacuum_index(Relation indrel, LVRelStats *vacrelstats)
 	vac_init_rusage(&ru0);
 
 	/*
-	 * If index is unsafe for concurrent access, must lock it.
+	 * Acquire appropriate type of lock on index: must be exclusive if
+	 * index AM isn't concurrent-safe.
 	 */
-	if (!indrel->rd_am->amconcurrent)
+	if (indrel->rd_am->amconcurrent)
+		LockRelation(indrel, RowExclusiveLock);
+	else
 		LockRelation(indrel, AccessExclusiveLock);
 
 	/* Do bulk deletion */
@@ -642,28 +684,34 @@ lazy_vacuum_index(Relation indrel, LVRelStats *vacrelstats)
 	/*
 	 * Release lock acquired above.
 	 */
-	if (!indrel->rd_am->amconcurrent)
+	if (indrel->rd_am->amconcurrent)
+		UnlockRelation(indrel, RowExclusiveLock);
+	else
 		UnlockRelation(indrel, AccessExclusiveLock);
 
 	if (!stats)
 		return;
 
+	/* accumulate total removed over multiple index-cleaning cycles */
+	*index_tups_vacuumed += stats->tuples_removed;
+	*index_pages_removed += stats->pages_removed;
+
 	/* now update statistics in pg_class */
-	vac_update_relstats(RelationGetRelid(indrel),
-						stats->num_pages, stats->num_index_tuples,
-						false);
+	lazy_update_relstats(indrel, stats->num_pages, *index_pages_removed,
+						 stats->num_index_tuples, *index_tups_vacuumed,
+						 false);
 
 	ereport(elevel,
-			(errmsg("index \"%s\" now contains %.0f row versions in %u pages",
-					RelationGetRelationName(indrel),
-					stats->num_index_tuples,
-					stats->num_pages),
-			 errdetail("%.0f index row versions were removed.\n"
+	   (errmsg("index \"%s\" now contains %.0f row versions in %u pages",
+			   RelationGetRelationName(indrel),
+			   stats->num_index_tuples,
+			   stats->num_pages),
+		errdetail("%.0f index row versions were removed.\n"
 		 "%u index pages have been deleted, %u are currently reusable.\n"
-					   "%s",
-					   stats->tuples_removed,
-					   stats->pages_deleted, stats->pages_free,
-					   vac_show_rusage(&ru0))));
+				  "%s",
+				  stats->tuples_removed,
+				  stats->pages_deleted, stats->pages_free,
+				  vac_show_rusage(&ru0))));
 
 	pfree(stats);
 }
@@ -731,18 +779,12 @@ lazy_truncate_heap(Relation onerel, LVRelStats *vacrelstats)
 	 * will also write out dirty buffers for blocks we aren't deleting,
 	 * but it's the closest thing in bufmgr's API.
 	 */
-	i = FlushRelationBuffers(onerel, new_rel_pages);
-	if (i < 0)
-		elog(ERROR, "FlushRelationBuffers returned %d", i);
+	FlushRelationBuffers(onerel, new_rel_pages);
 
 	/*
 	 * Do the physical truncation.
 	 */
-	new_rel_pages = smgrtruncate(DEFAULT_SMGR, onerel, new_rel_pages);
-	onerel->rd_nblocks = new_rel_pages; /* update relcache immediately */
-	onerel->rd_targblock = InvalidBlockNumber;
-	vacrelstats->rel_pages = new_rel_pages;		/* save new number of
-												 * blocks */
+	RelationTruncate(onerel, new_rel_pages);
 
 	/*
 	 * Drop free-space info for removed blocks; these must not get entered
@@ -763,6 +805,10 @@ lazy_truncate_heap(Relation onerel, LVRelStats *vacrelstats)
 	/* We destroyed the heap ordering, so mark array unordered */
 	vacrelstats->fs_is_heap = false;
 
+	/* update statistics */
+	vacrelstats->rel_pages = new_rel_pages;
+	vacrelstats->pages_removed = old_rel_pages - new_rel_pages;
+
 	/*
 	 * We keep the exclusive lock until commit (perhaps not necessary)?
 	 */
@@ -794,11 +840,10 @@ count_nondeletable_pages(Relation onerel, LVRelStats *vacrelstats)
 		Page		page;
 		OffsetNumber offnum,
 					maxoff;
-		bool		pgchanged,
-					tupgone,
+		bool		tupgone,
 					hastup;
 
-		CHECK_FOR_INTERRUPTS();
+		vacuum_delay_point();
 
 		blkno--;
 
@@ -811,13 +856,12 @@ count_nondeletable_pages(Relation onerel, LVRelStats *vacrelstats)
 
 		if (PageIsNew(page) || PageIsEmpty(page))
 		{
-			/* PageIsNew robably shouldn't happen... */
+			/* PageIsNew probably shouldn't happen... */
 			LockBuffer(buf, BUFFER_LOCK_UNLOCK);
 			ReleaseBuffer(buf);
 			continue;
 		}
 
-		pgchanged = false;
 		hastup = false;
 		maxoff = PageGetMaxOffsetNumber(page);
 		for (offnum = FirstOffsetNumber;
@@ -825,7 +869,6 @@ count_nondeletable_pages(Relation onerel, LVRelStats *vacrelstats)
 			 offnum = OffsetNumberNext(offnum))
 		{
 			ItemId		itemid;
-			uint16		sv_infomask;
 
 			itemid = PageGetItemId(page, offnum);
 
@@ -838,9 +881,8 @@ count_nondeletable_pages(Relation onerel, LVRelStats *vacrelstats)
 			ItemPointerSet(&(tuple.t_self), blkno, offnum);
 
 			tupgone = false;
-			sv_infomask = tuple.t_data->t_infomask;
 
-			switch (HeapTupleSatisfiesVacuum(tuple.t_data, OldestXmin))
+			switch (HeapTupleSatisfiesVacuum(tuple.t_data, OldestXmin, buf))
 			{
 				case HEAPTUPLE_DEAD:
 					tupgone = true;		/* we can delete the tuple */
@@ -866,10 +908,6 @@ count_nondeletable_pages(Relation onerel, LVRelStats *vacrelstats)
 					break;
 			}
 
-			/* check for hint-bit update by HeapTupleSatisfiesVacuum */
-			if (sv_infomask != tuple.t_data->t_infomask)
-				pgchanged = true;
-
 			if (!tupgone)
 			{
 				hastup = true;
@@ -879,10 +917,7 @@ count_nondeletable_pages(Relation onerel, LVRelStats *vacrelstats)
 
 		LockBuffer(buf, BUFFER_LOCK_UNLOCK);
 
-		if (pgchanged)
-			WriteBuffer(buf);
-		else
-			ReleaseBuffer(buf);
+		ReleaseBuffer(buf);
 
 		/* Done scanning if we found a tuple here */
 		if (hastup)
@@ -897,6 +932,51 @@ count_nondeletable_pages(Relation onerel, LVRelStats *vacrelstats)
 	return vacrelstats->nonempty_pages;
 }
 
+/*
+ * lazy_update_relstats - update pg_class statistics for a table or index
+ *
+ * We always want to set relpages to an accurate value.  However, for lazy
+ * VACUUM it seems best to set reltuples to the average of the number of
+ * rows before vacuuming and the number after vacuuming, rather than just
+ * using the number after vacuuming.  This will result in the best average
+ * performance in a steady-state situation where VACUUMs are performed
+ * regularly on a table of roughly constant size, assuming that the physical
+ * number of pages in the table stays about the same throughout.  (Note that
+ * we do not apply the same logic to VACUUM FULL, because it repacks the table
+ * and thereby boosts the tuple density.)
+ *
+ * An important point is that when the table size has decreased a lot during
+ * vacuuming, the old reltuples count might give an overestimate of the tuple
+ * density.  We handle this by scaling down the old reltuples count by the
+ * fraction by which the table has shortened before we merge it with the
+ * new reltuples count.  In particular this means that when relpages goes to
+ * zero, reltuples will immediately go to zero as well, causing the planner
+ * to fall back on other estimation procedures as the table grows again.
+ *
+ * Because we do this math independently for the table and the indexes, it's
+ * quite possible to end up with an index's reltuples different from the
+ * table's, which is silly except in the case of partial indexes.  We don't
+ * worry too much about that here; the planner contains filtering logic to
+ * ensure it only uses sane estimates.
+ */
+static void
+lazy_update_relstats(Relation rel, BlockNumber num_pages,
+					 BlockNumber pages_removed,
+					 double num_tuples, double tuples_removed,
+					 bool hasindex)
+{
+	double	old_num_tuples;
+
+	old_num_tuples = num_tuples + tuples_removed;
+	if (pages_removed > 0)
+		old_num_tuples *= (double) num_pages / (double) (num_pages + pages_removed);
+	if (old_num_tuples > num_tuples)
+		num_tuples = ceil((num_tuples + old_num_tuples) * 0.5);
+
+	vac_update_relstats(RelationGetRelid(rel), num_pages, num_tuples,
+						hasindex);
+}
+
 /*
  * lazy_space_alloc - space allocation decisions for lazy vacuum
  *
@@ -908,8 +988,8 @@ lazy_space_alloc(LVRelStats *vacrelstats, BlockNumber relblocks)
 	int			maxtuples;
 	int			maxpages;
 
-	maxtuples = (int) ((VacuumMem * 1024L) / sizeof(ItemPointerData));
-	/* stay sane if small VacuumMem */
+	maxtuples = (int) ((maintenance_work_mem * 1024L) / sizeof(ItemPointerData));
+	/* stay sane if small maintenance_work_mem */
 	if (maxtuples < MAX_TUPLES_PER_PAGE)
 		maxtuples = MAX_TUPLES_PER_PAGE;
 
@@ -922,9 +1002,6 @@ lazy_space_alloc(LVRelStats *vacrelstats, BlockNumber relblocks)
 	/* No need to allocate more pages than the relation has blocks */
 	if (relblocks < (BlockNumber) maxpages)
 		maxpages = (int) relblocks;
-	/* avoid palloc(0) */
-	if (maxpages < 1)
-		maxpages = 1;
 
 	vacrelstats->fs_is_heap = false;
 	vacrelstats->num_free_pages = 0;
@@ -942,8 +1019,8 @@ lazy_record_dead_tuple(LVRelStats *vacrelstats,
 {
 	/*
 	 * The array shouldn't overflow under normal behavior, but perhaps it
-	 * could if we are given a really small VacuumMem. In that case, just
-	 * forget the last few tuples.
+	 * could if we are given a really small maintenance_work_mem. In that
+	 * case, just forget the last few tuples.
 	 */
 	if (vacrelstats->num_dead_tuples < vacrelstats->max_dead_tuples)
 	{
