@@ -1,34 +1,34 @@
-/* -----------------------------------------------------------------------
+/*-----------------------------------------------------------------------
  * ascii.c
- *
- * $Header: /cvsroot/pgsql/src/backend/utils/adt/ascii.c,v 1.13.2.2 2003/07/14 16:41:46 tgl Exp $
- *
- *	 Portions Copyright (c) 1999-2000, PostgreSQL Global Development Group
- *
- *
- *	 TO_ASCII()
- *
  *	 The PostgreSQL routine for string to ascii conversion.
  *
- * -----------------------------------------------------------------------
+ *	 Portions Copyright (c) 1999-2005, PostgreSQL Global Development Group
+ *
+ * IDENTIFICATION
+ *	  $PostgreSQL: pgsql/src/backend/utils/adt/ascii.c,v 1.24 2005/01/01 05:43:07 momjian Exp $
+ *
+ *-----------------------------------------------------------------------
  */
-
 #include "postgres.h"
+
 #include "utils/builtins.h"
 #include "mb/pg_wchar.h"
 #include "utils/ascii.h"
 
+static void pg_to_ascii(unsigned char *src, unsigned char *src_end,
+			unsigned char *dest, int enc);
 static text *encode_to_ascii(text *data, int enc);
 
+
 /* ----------
  * to_ascii
  * ----------
  */
-char *
-pg_to_ascii(unsigned char *src, unsigned char *src_end, unsigned char *desc, int enc)
+static void
+pg_to_ascii(unsigned char *src, unsigned char *src_end, unsigned char *dest, int enc)
 {
 	unsigned char *x;
-	unsigned char *ascii;
+	const unsigned char *ascii;
 	int			range;
 
 	/*
@@ -37,7 +37,6 @@ pg_to_ascii(unsigned char *src, unsigned char *src_end, unsigned char *desc, int
 #define RANGE_128	128
 #define RANGE_160	160
 
-
 	if (enc == PG_LATIN1)
 	{
 		/*
@@ -54,6 +53,14 @@ pg_to_ascii(unsigned char *src, unsigned char *src_end, unsigned char *desc, int
 		ascii = " A L LS \"SSTZ-ZZ a,l'ls ,sstz\"zzRAAAALCCCEEEEIIDDNNOOOOxRUUUUYTBraaaalccceeeeiiddnnoooo/ruuuuyt.";
 		range = RANGE_160;
 	}
+	else if (enc == PG_LATIN9)
+	{
+		/*
+		 * ISO-8859-15 <range: 160 -- 255>
+		 */
+		ascii = "  cL YS sCa  -R     Zu .z   EeY?AAAAAAACEEEEIIII NOOOOOxOUUUUYTBaaaaaaaceeeeiiii nooooo/ouuuuyty";
+		range = RANGE_160;
+	}
 	else if (enc == PG_WIN1250)
 	{
 		/*
@@ -64,9 +71,11 @@ pg_to_ascii(unsigned char *src, unsigned char *src_end, unsigned char *desc, int
 	}
 	else
 	{
-		elog(ERROR, "pg_to_ascii(): unsupported encoding from %s",
-			 pg_encoding_to_char(enc));
-		return NULL;			/* keep compiler quiet */
+		ereport(ERROR,
+				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+			 errmsg("encoding conversion from %s to ASCII not supported",
+					pg_encoding_to_char(enc))));
+		return;					/* keep compiler quiet */
 	}
 
 	/*
@@ -75,27 +84,27 @@ pg_to_ascii(unsigned char *src, unsigned char *src_end, unsigned char *desc, int
 	for (x = src; x < src_end; x++)
 	{
 		if (*x < 128)
-			*desc++ = *x;
+			*dest++ = *x;
 		else if (*x < range)
-			*desc++ = ' ';		/* bogus 128 to 'range' */
+			*dest++ = ' ';		/* bogus 128 to 'range' */
 		else
-			*desc++ = ascii[*x - range];
+			*dest++ = ascii[*x - range];
 	}
-
-	return desc;
 }
 
 /* ----------
  * encode text
+ *
+ * The text datum is overwritten in-place, therefore this coding method
+ * cannot support conversions that change the string length!
  * ----------
  */
 static text *
 encode_to_ascii(text *data, int enc)
 {
-	pg_to_ascii(
-				(unsigned char *) VARDATA(data),		/* src */
-				(unsigned char *) (data) + VARSIZE(data),	/* src end */
-				(unsigned char *) VARDATA(data),		/* desc */
+	pg_to_ascii((unsigned char *) VARDATA(data),		/* src */
+				(unsigned char *) (data) + VARSIZE(data),		/* src end */
+				(unsigned char *) VARDATA(data),		/* dest */
 				enc);			/* encoding */
 
 	return data;
@@ -108,14 +117,10 @@ encode_to_ascii(text *data, int enc)
 Datum
 to_ascii_encname(PG_FUNCTION_ARGS)
 {
-	PG_RETURN_TEXT_P
-		(
-		 encode_to_ascii
-		 (
-		  PG_GETARG_TEXT_P_COPY(0),
-		  pg_char_to_encoding(NameStr(*PG_GETARG_NAME(1)))
-		  )
-		);
+	text	   *data = PG_GETARG_TEXT_P_COPY(0);
+	int			enc = pg_char_to_encoding(NameStr(*PG_GETARG_NAME(1)));
+
+	PG_RETURN_TEXT_P(encode_to_ascii(data, enc));
 }
 
 /* ----------
@@ -125,14 +130,10 @@ to_ascii_encname(PG_FUNCTION_ARGS)
 Datum
 to_ascii_enc(PG_FUNCTION_ARGS)
 {
-	PG_RETURN_TEXT_P
-		(
-		 encode_to_ascii
-		 (
-		  PG_GETARG_TEXT_P_COPY(0),
-		  PG_GETARG_INT32(1)
-		  )
-		);
+	text	   *data = PG_GETARG_TEXT_P_COPY(0);
+	int			enc = PG_GETARG_INT32(1);
+
+	PG_RETURN_TEXT_P(encode_to_ascii(data, enc));
 }
 
 /* ----------
@@ -142,12 +143,8 @@ to_ascii_enc(PG_FUNCTION_ARGS)
 Datum
 to_ascii_default(PG_FUNCTION_ARGS)
 {
-	PG_RETURN_TEXT_P
-		(
-		 encode_to_ascii
-		 (
-		  PG_GETARG_TEXT_P_COPY(0),
-		  GetDatabaseEncoding()
-		  )
-		);
+	text	   *data = PG_GETARG_TEXT_P_COPY(0);
+	int			enc = GetDatabaseEncoding();
+
+	PG_RETURN_TEXT_P(encode_to_ascii(data, enc));
 }
