@@ -3,14 +3,14 @@
  * fe-auth.c
  *	   The front-end (client) authorization routines
  *
- * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  * NOTE: the error message strings returned by this module must not
  * exceed INITIAL_EXPBUFFER_SIZE (currently 256 bytes).
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/interfaces/libpq/fe-auth.c,v 1.84.2.4 2007/07/23 18:10:13 mha Exp $
+ *	  $PostgreSQL: pgsql/src/interfaces/libpq/fe-auth.c,v 1.99 2005/01/12 21:37:54 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -170,10 +170,10 @@ pg_krb4_authname(char *PQerrormsg)
 	status = krb_get_tf_fullname(tkt_string(), name, instance, realm);
 	if (status != KSUCCESS)
 	{
-		snprintf(PQerrormsg, INITIAL_EXPBUFFER_SIZE,
+		snprintf(PQerrormsg, PQERRORMSG_LENGTH,
 				 "pg_krb4_authname: krb_get_tf_fullname: %s\n",
 				 krb_err_txt[status]);
-		return (char *) NULL;
+		return NULL;
 	}
 	return name;
 }
@@ -220,15 +220,15 @@ pg_krb4_sendauth(char *PQerrormsg, int sock,
 						  hostname,
 						  realm,
 						  (u_long) 0,
-						  (MSG_DAT *) NULL,
-						  (CREDENTIALS *) NULL,
+						  NULL,
+						  NULL,
 						  NULL,
 						  laddr,
 						  raddr,
 						  PG_KRB4_VERSION);
 	if (status != KSUCCESS)
 	{
-		snprintf(PQerrormsg, INITIAL_EXPBUFFER_SIZE,
+		snprintf(PQerrormsg, PQERRORMSG_LENGTH,
 				 libpq_gettext("Kerberos 4 error: %s\n"),
 				 krb_err_txt[status]);
 		return STATUS_ERROR;
@@ -243,7 +243,10 @@ pg_krb4_sendauth(char *PQerrormsg, int sock,
  */
 
 #include <krb5.h>
+/* Some old versions of Kerberos do not include <com_err.h> in <krb5.h> */
+#if !defined(__COM_ERR_H) && !defined(__COM_ERR_H__)
 #include <com_err.h>
+#endif
 
 /*
  * pg_an_to_ln -- return the local name corresponding to an authentication
@@ -291,7 +294,7 @@ pg_krb5_init(char *PQerrormsg)
 	retval = krb5_init_context(&pg_krb5_context);
 	if (retval)
 	{
-		snprintf(PQerrormsg, INITIAL_EXPBUFFER_SIZE,
+		snprintf(PQerrormsg, PQERRORMSG_LENGTH,
 				 "pg_krb5_init: krb5_init_context: %s\n",
 				 error_message(retval));
 		return STATUS_ERROR;
@@ -300,7 +303,7 @@ pg_krb5_init(char *PQerrormsg)
 	retval = krb5_cc_default(pg_krb5_context, &pg_krb5_ccache);
 	if (retval)
 	{
-		snprintf(PQerrormsg, INITIAL_EXPBUFFER_SIZE,
+		snprintf(PQerrormsg, PQERRORMSG_LENGTH,
 				 "pg_krb5_init: krb5_cc_default: %s\n",
 				 error_message(retval));
 		krb5_free_context(pg_krb5_context);
@@ -311,7 +314,7 @@ pg_krb5_init(char *PQerrormsg)
 								   &pg_krb5_client);
 	if (retval)
 	{
-		snprintf(PQerrormsg, INITIAL_EXPBUFFER_SIZE,
+		snprintf(PQerrormsg, PQERRORMSG_LENGTH,
 				 "pg_krb5_init: krb5_cc_get_principal: %s\n",
 				 error_message(retval));
 		krb5_cc_close(pg_krb5_context, pg_krb5_ccache);
@@ -322,7 +325,7 @@ pg_krb5_init(char *PQerrormsg)
 	retval = krb5_unparse_name(pg_krb5_context, pg_krb5_client, &pg_krb5_name);
 	if (retval)
 	{
-		snprintf(PQerrormsg, INITIAL_EXPBUFFER_SIZE,
+		snprintf(PQerrormsg, PQERRORMSG_LENGTH,
 				 "pg_krb5_init: krb5_unparse_name: %s\n",
 				 error_message(retval));
 		krb5_free_principal(pg_krb5_context, pg_krb5_client);
@@ -374,7 +377,7 @@ pg_krb5_sendauth(char *PQerrormsg, int sock, const char *hostname)
 									 KRB5_NT_SRV_HST, &server);
 	if (retval)
 	{
-		snprintf(PQerrormsg, INITIAL_EXPBUFFER_SIZE,
+		snprintf(PQerrormsg, PQERRORMSG_LENGTH,
 				 "pg_krb5_sendauth: krb5_sname_to_principal: %s\n",
 				 error_message(retval));
 		return STATUS_ERROR;
@@ -390,7 +393,7 @@ pg_krb5_sendauth(char *PQerrormsg, int sock, const char *hostname)
 	{
 		char		sebuf[256];
 
-		snprintf(PQerrormsg, INITIAL_EXPBUFFER_SIZE,
+		snprintf(PQerrormsg, PQERRORMSG_LENGTH,
 				 libpq_gettext("could not set socket to blocking mode: %s\n"), pqStrerror(errno, sebuf, sizeof(sebuf)));
 		krb5_free_principal(pg_krb5_context, server);
 		return STATUS_ERROR;
@@ -407,11 +410,11 @@ pg_krb5_sendauth(char *PQerrormsg, int sock, const char *hostname)
 		if (retval == KRB5_SENDAUTH_REJECTED && err_ret)
 		{
 #if defined(HAVE_KRB5_ERROR_TEXT_DATA)
-			snprintf(PQerrormsg, INITIAL_EXPBUFFER_SIZE,
+			snprintf(PQerrormsg, PQERRORMSG_LENGTH,
 			  libpq_gettext("Kerberos 5 authentication rejected: %*s\n"),
 					 (int) err_ret->text.length, err_ret->text.data);
 #elif defined(HAVE_KRB5_ERROR_E_DATA)
-			snprintf(PQerrormsg, INITIAL_EXPBUFFER_SIZE,
+			snprintf(PQerrormsg, PQERRORMSG_LENGTH,
 			  libpq_gettext("Kerberos 5 authentication rejected: %*s\n"),
 					 (int) err_ret->e_data->length,
 					 (const char *) err_ret->e_data->data);
@@ -421,7 +424,7 @@ pg_krb5_sendauth(char *PQerrormsg, int sock, const char *hostname)
 		}
 		else
 		{
-			snprintf(PQerrormsg, INITIAL_EXPBUFFER_SIZE,
+			snprintf(PQerrormsg, PQERRORMSG_LENGTH,
 					 "krb5_sendauth: %s\n", error_message(retval));
 		}
 
@@ -437,7 +440,7 @@ pg_krb5_sendauth(char *PQerrormsg, int sock, const char *hostname)
 	{
 		char		sebuf[256];
 
-		snprintf(PQerrormsg, INITIAL_EXPBUFFER_SIZE,
+		snprintf(PQerrormsg, PQERRORMSG_LENGTH,
 				 libpq_gettext("could not restore non-blocking mode on socket: %s\n"),
 				 pqStrerror(errno, sebuf, sizeof(sebuf)));
 		ret = STATUS_ERROR;
@@ -497,14 +500,14 @@ pg_local_sendauth(char *PQerrormsg, PGconn *conn)
 	{
 		char		sebuf[256];
 
-		snprintf(PQerrormsg, INITIAL_EXPBUFFER_SIZE,
+		snprintf(PQerrormsg, PQERRORMSG_LENGTH,
 				 "pg_local_sendauth: sendmsg: %s\n",
 				 pqStrerror(errno, sebuf, sizeof(sebuf)));
 		return STATUS_ERROR;
 	}
 	return STATUS_OK;
 #else
-	snprintf(PQerrormsg, INITIAL_EXPBUFFER_SIZE,
+	snprintf(PQerrormsg, PQERRORMSG_LENGTH,
 		libpq_gettext("SCM_CRED authentication method not supported\n"));
 	return STATUS_ERROR;
 #endif
@@ -527,7 +530,7 @@ pg_password_sendauth(PGconn *conn, const char *password, AuthRequest areq)
 				if (!(crypt_pwd = malloc(MD5_PASSWD_LEN + 1)) ||
 					!(crypt_pwd2 = malloc(MD5_PASSWD_LEN + 1)))
 				{
-					perror("malloc");
+					fprintf(stderr, libpq_gettext("out of memory\n"));
 					return STATUS_ERROR;
 				}
 				if (!EncryptMD5(password, conn->pguser,
@@ -590,34 +593,40 @@ fe_sendauth(AuthRequest areq, PGconn *conn, const char *hostname,
 
 		case AUTH_REQ_KRB4:
 #ifdef KRB4
+			pglock_thread();
 			if (pg_krb4_sendauth(PQerrormsg, conn->sock,
 							   (struct sockaddr_in *) & conn->laddr.addr,
 							   (struct sockaddr_in *) & conn->raddr.addr,
 								 hostname) != STATUS_OK)
 			{
-				snprintf(PQerrormsg, INITIAL_EXPBUFFER_SIZE,
+				snprintf(PQerrormsg, PQERRORMSG_LENGTH,
 					libpq_gettext("Kerberos 4 authentication failed\n"));
+				pgunlock_thread();
 				return STATUS_ERROR;
 			}
+			pgunlock_thread();
 			break;
 #else
-			snprintf(PQerrormsg, INITIAL_EXPBUFFER_SIZE,
+			snprintf(PQerrormsg, PQERRORMSG_LENGTH,
 			 libpq_gettext("Kerberos 4 authentication not supported\n"));
 			return STATUS_ERROR;
 #endif
 
 		case AUTH_REQ_KRB5:
 #ifdef KRB5
+			pglock_thread();
 			if (pg_krb5_sendauth(PQerrormsg, conn->sock,
 								 hostname) != STATUS_OK)
 			{
-				snprintf(PQerrormsg, INITIAL_EXPBUFFER_SIZE,
+				snprintf(PQerrormsg, PQERRORMSG_LENGTH,
 					libpq_gettext("Kerberos 5 authentication failed\n"));
+				pgunlock_thread();
 				return STATUS_ERROR;
 			}
+			pgunlock_thread();
 			break;
 #else
-			snprintf(PQerrormsg, INITIAL_EXPBUFFER_SIZE,
+			snprintf(PQerrormsg, PQERRORMSG_LENGTH,
 			 libpq_gettext("Kerberos 5 authentication not supported\n"));
 			return STATUS_ERROR;
 #endif
@@ -627,13 +636,13 @@ fe_sendauth(AuthRequest areq, PGconn *conn, const char *hostname,
 		case AUTH_REQ_PASSWORD:
 			if (password == NULL || *password == '\0')
 			{
-				(void) snprintf(PQerrormsg, INITIAL_EXPBUFFER_SIZE,
-								"fe_sendauth: no password supplied\n");
+				(void) snprintf(PQerrormsg, PQERRORMSG_LENGTH,
+								PQnoPasswordSupplied);
 				return STATUS_ERROR;
 			}
 			if (pg_password_sendauth(conn, password, areq) != STATUS_OK)
 			{
-				(void) snprintf(PQerrormsg, INITIAL_EXPBUFFER_SIZE,
+				(void) snprintf(PQerrormsg, PQERRORMSG_LENGTH,
 				 "fe_sendauth: error sending password authentication\n");
 				return STATUS_ERROR;
 			}
@@ -645,7 +654,7 @@ fe_sendauth(AuthRequest areq, PGconn *conn, const char *hostname,
 			break;
 
 		default:
-			snprintf(PQerrormsg, INITIAL_EXPBUFFER_SIZE,
+			snprintf(PQerrormsg, PQERRORMSG_LENGTH,
 					 libpq_gettext("authentication method %u not supported\n"), areq);
 			return STATUS_ERROR;
 	}
@@ -682,7 +691,7 @@ fe_setauthsvc(const char *name, char *PQerrormsg)
 		}
 	if (i == n_authsvcs)
 	{
-		snprintf(PQerrormsg, INITIAL_EXPBUFFER_SIZE,
+		snprintf(PQerrormsg, PQERRORMSG_LENGTH,
 				 libpq_gettext("invalid authentication service name \"%s\", ignored\n"),
 				 name);
 	}
@@ -712,9 +721,17 @@ fe_getauthsvc(char *PQerrormsg)
 char *
 fe_getauthname(char *PQerrormsg)
 {
-	const char *name = (char *) NULL;
-	char	   *authn = (char *) NULL;
+	const char *name = NULL;
+	char	   *authn;
 	MsgType		authsvc;
+#ifdef WIN32
+	char		username[128];
+	DWORD		namesize = sizeof(username) - 1;
+#else
+	char		pwdbuf[BUFSIZ];
+	struct passwd pwdstr;
+	struct passwd *pw = NULL;
+#endif
 
 	authsvc = fe_getauthsvc(PQerrormsg);
 
@@ -722,6 +739,8 @@ fe_getauthname(char *PQerrormsg)
 	if (authsvc == 0)
 		return NULL;			/* leave original error message in place */
 
+	pglock_thread();
+
 #ifdef KRB4
 	if (authsvc == STARTUP_KRB4_MSG)
 		name = pg_krb4_authname(PQerrormsg);
@@ -736,28 +755,22 @@ fe_getauthname(char *PQerrormsg)
 		|| (authsvc == STARTUP_KRB5_MSG && !name))
 	{
 #ifdef WIN32
-		char		username[128];
-		DWORD		namesize = sizeof(username) - 1;
-
 		if (GetUserName(username, &namesize))
 			name = username;
 #else
-		char		pwdbuf[BUFSIZ];
-		struct passwd pwdstr;
-		struct passwd *pw = NULL;
-
-		if (pqGetpwuid(geteuid(), &pwdstr,
-					   pwdbuf, sizeof(pwdbuf), &pw) == 0)
+		if (pqGetpwuid(geteuid(), &pwdstr, pwdbuf, sizeof(pwdbuf), &pw) == 0)
 			name = pw->pw_name;
 #endif
 	}
 
 	if (authsvc != STARTUP_MSG && authsvc != STARTUP_KRB4_MSG && authsvc != STARTUP_KRB5_MSG)
-		snprintf(PQerrormsg, INITIAL_EXPBUFFER_SIZE,
+		snprintf(PQerrormsg, PQERRORMSG_LENGTH,
 				 libpq_gettext("fe_getauthname: invalid authentication system: %d\n"),
 				 authsvc);
 
-	if (name && (authn = (char *) malloc(strlen(name) + 1)))
-		strcpy(authn, name);
+	authn = name ? strdup(name) : NULL;
+
+	pgunlock_thread();
+
 	return authn;
 }
