@@ -3,12 +3,12 @@
  * fe-connect.c
  *	  functions related to setting up a connection to the backend
  *
- * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/interfaces/libpq/fe-connect.c,v 1.263.2.4 2005/07/14 14:07:41 tgl Exp $
+ *	  $PostgreSQL: pgsql/src/interfaces/libpq/fe-connect.c,v 1.301 2005/01/14 00:25:56 momjian Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -18,9 +18,11 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <fcntl.h>
-#include <errno.h>
 #include <ctype.h>
 #include <time.h>
+#ifndef WIN32_CLIENT_ONLY
+#include <unistd.h>
+#endif
 
 #ifndef HAVE_STRDUP
 #include "strdup.h"
@@ -29,12 +31,21 @@
 #include "libpq-fe.h"
 #include "libpq-int.h"
 #include "fe-auth.h"
+#include "pg_config_paths.h"
 
 #ifdef WIN32
 #include "win32.h"
+#ifdef _WIN32_IE
+#undef _WIN32_IE
+#endif
+#define _WIN32_IE 0x0400
+#ifdef near
+#undef near
+#endif
+#define near
+#include <shlobj.h>
 #else
 #include <sys/socket.h>
-#include <unistd.h>
 #include <netdb.h>
 #include <netinet/in.h>
 #ifdef HAVE_NETINET_TCP_H
@@ -43,15 +54,23 @@
 #include <arpa/inet.h>
 #endif
 
+#ifdef ENABLE_THREAD_SAFETY
+#include <pthread.h>
+#endif
+
 #include "libpq/ip.h"
 #include "mb/pg_wchar.h"
 
-/* For FNCTL_NONBLOCK */
-#if defined(WIN32) || defined(__BEOS__)
-long		ioctlsocket_ret=1;
+#ifndef FD_CLOEXEC
+#define FD_CLOEXEC 1
 #endif
 
+
+#ifndef WIN32
 #define PGPASSFILE ".pgpass"
+#else
+#define PGPASSFILE "pgpass.conf"
+#endif
 
 /* fall back options if they are not specified by arguments or defined
    by environment variables */
@@ -66,7 +85,6 @@ long		ioctlsocket_ret=1;
 #define DefaultSSLMode	"disable"
 #endif
 
-
 /* ----------
  * Definition of the conninfo parameters and their fallback resources.
  *
@@ -197,6 +215,12 @@ static int parseServiceInfo(PQconninfoOption *options,
 static char *pwdfMatchesString(char *buf, char *token);
 static char *PasswordFromFile(char *hostname, char *port, char *dbname,
 				 char *username);
+static void default_threadlock(int acquire);
+
+
+/* global variable because fe-auth.c needs to access it */
+pgthreadlock_t pg_g_threadlock = default_threadlock;
+
 
 /*
  *		Connecting to a Database
@@ -279,7 +303,7 @@ PQconnectStart(const char *conninfo)
 	 */
 	conn = makeEmptyPGconn();
 	if (conn == NULL)
-		return (PGconn *) NULL;
+		return NULL;
 
 	/*
 	 * Parse the conninfo string
@@ -449,7 +473,7 @@ connectOptions2(PGconn *conn)
 		{
 			conn->status = CONNECTION_BAD;
 			printfPQExpBuffer(&conn->errorMessage,
-						 libpq_gettext("invalid sslmode value: \"%s\"\n"),
+						libpq_gettext("invalid sslmode value: \"%s\"\n"),
 							  conn->sslmode);
 			return false;
 		}
@@ -532,7 +556,7 @@ PQsetdbLogin(const char *pghost, const char *pgport, const char *pgoptions,
 	 */
 	conn = makeEmptyPGconn();
 	if (conn == NULL)
-		return (PGconn *) NULL;
+		return NULL;
 
 	/*
 	 * Parse an empty conninfo string in order to set up the same defaults
@@ -679,7 +703,7 @@ update_db_info(PGconn *conn)
 				*tmp = '\0';
 			}
 
-			tmp = last_path_separator(conn->dbName + offset);
+			tmp = last_dir_separator(conn->dbName + offset);
 			if (tmp != NULL)	/* database name given */
 			{
 				if (conn->dbName)
@@ -766,28 +790,6 @@ update_db_info(PGconn *conn)
 #endif   /* NOT_USED */
 
 
-/* ----------
- * connectMakeNonblocking -
- * Make a connection non-blocking.
- * Returns 1 if successful, 0 if not.
- * ----------
- */
-static int
-connectMakeNonblocking(PGconn *conn)
-{
-	if (FCNTL_NONBLOCK(conn->sock) < 0)
-	{
-		char		sebuf[256];
-
-		printfPQExpBuffer(&conn->errorMessage,
-		libpq_gettext("could not set socket to non-blocking mode: %s\n"),
-						SOCK_STRERROR(SOCK_ERRNO, sebuf, sizeof(sebuf)));
-		return 0;
-	}
-
-	return 1;
-}
-
 /* ----------
  * connectNoDelay -
  * Sets the TCP_NODELAY socket option.
@@ -879,7 +881,7 @@ connectDBStart(PGconn *conn)
 	char		portstr[128];
 	struct addrinfo *addrs = NULL;
 	struct addrinfo hint;
-	const char *node = NULL;
+	const char *node;
 	int			ret;
 
 	if (!conn)
@@ -918,15 +920,19 @@ connectDBStart(PGconn *conn)
 		node = conn->pghost;
 		hint.ai_family = AF_UNSPEC;
 	}
-#ifdef HAVE_UNIX_SOCKETS
 	else
 	{
+#ifdef HAVE_UNIX_SOCKETS
 		/* pghostaddr and pghost are NULL, so use Unix domain socket */
 		node = NULL;
 		hint.ai_family = AF_UNIX;
 		UNIXSOCK_PATH(portstr, portnum, conn->pgunixsocket);
-	}
+#else
+		/* Without Unix sockets, default to localhost instead */
+		node = "localhost";
+		hint.ai_family = AF_UNSPEC;
 #endif   /* HAVE_UNIX_SOCKETS */
+	}
 
 	/* Use getaddrinfo_all() to resolve the address */
 	ret = getaddrinfo_all(node, portstr, &hint, &addrs);
@@ -940,7 +946,8 @@ connectDBStart(PGconn *conn)
 			printfPQExpBuffer(&conn->errorMessage,
 							  libpq_gettext("could not translate Unix-domain socket path \"%s\" to address: %s\n"),
 							  portstr, gai_strerror(ret));
-		freeaddrinfo_all(hint.ai_family, addrs);
+		if (addrs)
+			freeaddrinfo_all(hint.ai_family, addrs);
 		goto connect_errReturn;
 	}
 
@@ -1191,8 +1198,8 @@ keep_going:						/* We will come back to here until there
 
 					/*
 					 * Select socket options: no delay of outgoing data
-					 * for TCP sockets, and nonblock mode.	Fail if this
-					 * fails.
+					 * for TCP sockets, nonblock mode, close-on-exec.
+					 * Fail if any of this fails.
 					 */
 					if (!IS_AF_UNIX(addr_cur->ai_family))
 					{
@@ -1204,14 +1211,30 @@ keep_going:						/* We will come back to here until there
 							continue;
 						}
 					}
-					if (connectMakeNonblocking(conn) == 0)
+					if (!set_noblock(conn->sock))
 					{
+						printfPQExpBuffer(&conn->errorMessage,
+										  libpq_gettext("could not set socket to non-blocking mode: %s\n"),
+										  SOCK_STRERROR(SOCK_ERRNO, sebuf, sizeof(sebuf)));
 						closesocket(conn->sock);
 						conn->sock = -1;
 						conn->addr_cur = addr_cur->ai_next;
 						continue;
 					}
 
+#ifdef F_SETFD
+					if (fcntl(conn->sock, F_SETFD, FD_CLOEXEC) == -1)
+					{
+						printfPQExpBuffer(&conn->errorMessage,
+										  libpq_gettext("could not set socket to close-on-exec mode: %s\n"),
+										  SOCK_STRERROR(SOCK_ERRNO, sebuf, sizeof(sebuf)));
+						closesocket(conn->sock);
+						conn->sock = -1;
+						conn->addr_cur = addr_cur->ai_next;
+						continue;
+					}
+#endif /* F_SETFD */
+
 					/*
 					 * Start/make connection.  This should not block,
 					 * since we are in nonblock mode.  If it does, well,
@@ -1437,30 +1460,36 @@ keep_going:						/* We will come back to here until there
 
 				/*
 				 * On first time through, get the postmaster's response to
-				 * our SSL negotiation packet.	Be careful to read only
-				 * one byte (if there's more, it could be SSL data).
+				 * our SSL negotiation packet.
 				 */
 				if (conn->ssl == NULL)
 				{
+					/*
+					 * We use pqReadData here since it has the logic to
+					 * distinguish no-data-yet from connection closure.
+					 * Since conn->ssl isn't set, a plain recv() will occur.
+					 */
 					char		SSLok;
-					int			nread;
+					int			rdresult;
 
-			retry_ssl_read:
-					nread = recv(conn->sock, &SSLok, 1, 0);
-					if (nread < 0)
+					rdresult = pqReadData(conn);
+					if (rdresult < 0)
 					{
-						if (SOCK_ERRNO == EINTR)
-							/* Interrupted system call - just try again */
-							goto retry_ssl_read;
-
-						printfPQExpBuffer(&conn->errorMessage,
-										  libpq_gettext("could not receive server response to SSL negotiation packet: %s\n"),
-						SOCK_STRERROR(SOCK_ERRNO, sebuf, sizeof(sebuf)));
+						/* errorMessage is already filled in */
 						goto error_return;
 					}
-					if (nread == 0)
+					if (rdresult == 0)
+					{
 						/* caller failed to wait for data */
 						return PGRES_POLLING_READING;
+					}
+					if (pqGetc(&SSLok, conn) < 0)
+					{
+						/* should not happen really */
+						return PGRES_POLLING_READING;
+					}
+					/* mark byte consumed */
+					conn->inStart = conn->inCursor;
 					if (SSLok == 'S')
 					{
 						/* Do one-time setup; this creates conn->ssl */
@@ -1934,11 +1963,11 @@ makeEmptyPGconn(void)
 	PGconn	   *conn;
 
 #ifdef WIN32
-	/* needed to use the static libpq under windows as well */
+	/* make sure socket support is up and running */
 	WSADATA		wsaData;
 
 	if (WSAStartup(MAKEWORD(1, 1), &wsaData))
-		return (PGconn *) NULL;
+		return NULL;
 	WSASetLastError(0);
 #endif
 
@@ -1947,7 +1976,7 @@ makeEmptyPGconn(void)
 		return conn;
 
 	/* Zero all pointers and booleans */
-	MemSet((char *) conn, 0, sizeof(PGconn));
+	MemSet(conn, 0, sizeof(PGconn));
 
 	conn->noticeHooks.noticeRec = defaultNoticeReceiver;
 	conn->noticeHooks.noticeProc = defaultNoticeProcessor;
@@ -1957,7 +1986,6 @@ makeEmptyPGconn(void)
 	conn->setenv_state = SETENV_STATE_IDLE;
 	conn->client_encoding = PG_SQL_ASCII;
 	conn->verbosity = PQERRORS_DEFAULT;
-	conn->notifyList = DLNewList();
 	conn->sock = -1;
 #ifdef USE_SSL
 	conn->allow_ssl_try = true;
@@ -1999,18 +2027,13 @@ makeEmptyPGconn(void)
 /*
  * freePGconn
  *	 - free the PGconn data structure
- *
- * When changing/adding to this function, see also closePGconn!
  */
 static void
 freePGconn(PGconn *conn)
 {
+	PGnotify   *notify;
 	pgParameterStatus *pstatus;
 
-#ifdef WIN32
-	WSACleanup();
-#endif
-
 	if (!conn)
 		return;
 
@@ -2043,9 +2066,15 @@ freePGconn(PGconn *conn)
 	if (conn->sslmode)
 		free(conn->sslmode);
 	/* Note that conn->Pfdebug is not ours to close or free */
+	notify = conn->notifyHead;
+	while (notify != NULL)
+	{
+		PGnotify *prev = notify;
+
+		notify = notify->next;
+		free(prev);
+	}
 	freeaddrinfo_all(conn->addrlist_family, conn->addrlist);
-	if (conn->notifyList)
-		DLFreeList(conn->notifyList);
 	pstatus = conn->pstatus;
 	while (pstatus != NULL)
 	{
@@ -2068,14 +2097,10 @@ freePGconn(PGconn *conn)
 /*
  * closePGconn
  *	 - properly close a connection to the backend
- *
- * Release all transient state, but NOT the connection parameters.
  */
 static void
 closePGconn(PGconn *conn)
 {
-	pgParameterStatus *pstatus;
-
 	/*
 	 * Note that the protocol doesn't allow us to send Terminate messages
 	 * during the startup phase.
@@ -2111,21 +2136,6 @@ closePGconn(PGconn *conn)
 										 * absent */
 	conn->asyncStatus = PGASYNC_IDLE;
 	pqClearAsyncResult(conn);	/* deallocate result and curTuple */
-	freeaddrinfo_all(conn->addrlist_family, conn->addrlist);
-	conn->addrlist = NULL;
-	conn->addr_cur = NULL;
-	if (conn->notifyList)
-		DLFreeList(conn->notifyList);
-	conn->notifyList = DLNewList();
-	pstatus = conn->pstatus;
-	while (pstatus != NULL)
-	{
-		pgParameterStatus *prev = pstatus;
-
-		pstatus = pstatus->next;
-		free(prev);
-	}
-	conn->pstatus = NULL;
 	if (conn->lobjfuncs)
 		free(conn->lobjfuncs);
 	conn->lobjfuncs = NULL;
@@ -2199,19 +2209,55 @@ PQresetPoll(PGconn *conn)
 	return PGRES_POLLING_FAILED;
 }
 
+/*
+ * PQcancelGet: get a PGcancel structure corresponding to a connection.
+ *
+ * A copy is needed to be able to cancel a running query from a different
+ * thread. If the same structure is used all structure members would have
+ * to be individually locked (if the entire structure was locked, it would
+ * be impossible to cancel a synchronous query becuase the structure would
+ * have to stay locked for the duration of the query).
+ */
+PGcancel *
+PQgetCancel(PGconn *conn)
+{
+	PGcancel *cancel;
+
+	if (!conn)
+		return NULL;
+
+	if (conn->sock < 0)
+		return NULL;
+
+	cancel = malloc(sizeof(PGcancel));
+	if (cancel == NULL)
+		return NULL;
+
+	memcpy(&cancel->raddr, &conn->raddr, sizeof(SockAddr));
+	cancel->be_pid = conn->be_pid;
+	cancel->be_key = conn->be_key;
+
+	return cancel;
+}
+
+/* PQfreeCancel: free a cancel structure */
+void
+PQfreeCancel(PGcancel *cancel)
+{
+	if (cancel)
+		free(cancel);
+}
+
 
 /*
- * PQrequestCancel: attempt to request cancellation of the current operation.
+ * PQcancel and PQrequestCancel: attempt to request cancellation of the
+ * current operation.
  *
  * The return value is TRUE if the cancel request was successfully
- * dispatched, FALSE if not (in which case conn->errorMessage is set).
+ * dispatched, FALSE if not (in which case an error message is available).
  * Note: successful dispatch is no guarantee that there will be any effect at
  * the backend.  The application must read the operation result as usual.
  *
- * XXX it was a bad idea to have the error message returned in
- * conn->errorMessage, since it could overwrite a message already there.
- * Would be better to return it in a char array passed by the caller.
- *
  * CAUTION: we want this routine to be safely callable from a signal handler
  * (for example, an application might want to call it in a SIGINT handler).
  * This means we cannot use any C library routine that might be non-reentrant.
@@ -2220,59 +2266,40 @@ PQresetPoll(PGconn *conn)
  * error messages with strcpy/strcat is tedious but should be quite safe.
  * We also save/restore errno in case the signal handler support doesn't.
  *
- * NOTE: this routine must not generate any error message longer than
- * INITIAL_EXPBUFFER_SIZE (currently 256), since we dare not try to
- * expand conn->errorMessage!
+ * internal_cancel() is an internal helper function to make code-sharing
+ * between the two versions of the cancel function possible.
  */
-
-int
-PQrequestCancel(PGconn *conn)
+static int
+internal_cancel(SockAddr *raddr, int be_pid, int be_key,
+				char *errbuf, int errbufsize)
 {
 	int			save_errno = SOCK_ERRNO;
 	int			tmpsock = -1;
 	char		sebuf[256];
+	int         maxlen;
 	struct
 	{
 		uint32		packetlen;
 		CancelRequestPacket cp;
 	}			crp;
 
-	/* Check we have an open connection */
-	if (!conn)
-		return FALSE;
-
-	if (conn->sock < 0)
-	{
-		strcpy(conn->errorMessage.data,
-			   "PQrequestCancel() -- connection is not open\n");
-		conn->errorMessage.len = strlen(conn->errorMessage.data);
-#ifdef WIN32
-		WSASetLastError(save_errno);
-#else
-		errno = save_errno;
-#endif
-		return FALSE;
-	}
-
 	/*
-	 * We need to open a temporary connection to the postmaster. Use the
-	 * information saved by connectDB to do this with only kernel calls.
+	 * We need to open a temporary connection to the postmaster. Do
+	 * this with only kernel calls.
 	 */
-	if ((tmpsock = socket(conn->raddr.addr.ss_family, SOCK_STREAM, 0)) < 0)
+	if ((tmpsock = socket(raddr->addr.ss_family, SOCK_STREAM, 0)) < 0)
 	{
-		strcpy(conn->errorMessage.data,
-			   "PQrequestCancel() -- socket() failed: ");
+		StrNCpy(errbuf, "PQcancel() -- socket() failed: ", errbufsize);
 		goto cancel_errReturn;
 	}
 retry3:
-	if (connect(tmpsock, (struct sockaddr *) & conn->raddr.addr,
-				conn->raddr.salen) < 0)
+	if (connect(tmpsock, (struct sockaddr *) & raddr->addr,
+				raddr->salen) < 0)
 	{
 		if (SOCK_ERRNO == EINTR)
 			/* Interrupted system call - we'll just try again */
 			goto retry3;
-		strcpy(conn->errorMessage.data,
-			   "PQrequestCancel() -- connect() failed: ");
+		StrNCpy(errbuf, "PQcancel() -- connect() failed: ", errbufsize);
 		goto cancel_errReturn;
 	}
 
@@ -2284,8 +2311,8 @@ retry3:
 
 	crp.packetlen = htonl((uint32) sizeof(crp));
 	crp.cp.cancelRequestCode = (MsgType) htonl(CANCEL_REQUEST_CODE);
-	crp.cp.backendPID = htonl(conn->be_pid);
-	crp.cp.cancelAuthCode = htonl(conn->be_key);
+	crp.cp.backendPID = htonl(be_pid);
+	crp.cp.cancelAuthCode = htonl(be_key);
 
 retry4:
 	if (send(tmpsock, (char *) &crp, sizeof(crp), 0) != (int) sizeof(crp))
@@ -2293,17 +2320,17 @@ retry4:
 		if (SOCK_ERRNO == EINTR)
 			/* Interrupted system call - we'll just try again */
 			goto retry4;
-		strcpy(conn->errorMessage.data,
-			   "PQrequestCancel() -- send() failed: ");
+		StrNCpy(errbuf, "PQcancel() -- send() failed: ", errbufsize);
 		goto cancel_errReturn;
 	}
 
 	/*
-	 * Wait for the postmaster to close the connection, which indicates that
-	 * it's processed the request.  Without this delay, we might issue another
-	 * command only to find that our cancel zaps that command instead of the
-	 * one we thought we were canceling.  Note we don't actually expect this
-	 * read to obtain any data, we are just waiting for EOF to be signaled.
+	 * Wait for the postmaster to close the connection, which indicates
+	 * that it's processed the request.  Without this delay, we might
+	 * issue another command only to find that our cancel zaps that
+	 * command instead of the one we thought we were canceling.  Note we
+	 * don't actually expect this read to obtain any data, we are just
+	 * waiting for EOF to be signaled.
 	 */
 retry5:
 	if (recv(tmpsock, (char *) &crp, 1, 0) < 0)
@@ -2316,30 +2343,89 @@ retry5:
 
 	/* All done */
 	closesocket(tmpsock);
-#ifdef WIN32
-	WSASetLastError(save_errno);
-#else
-	errno = save_errno;
-#endif
-
+	SOCK_ERRNO_SET(save_errno);
 	return TRUE;
 
 cancel_errReturn:
-	strcat(conn->errorMessage.data, SOCK_STRERROR(SOCK_ERRNO, sebuf, sizeof(sebuf)));
-	strcat(conn->errorMessage.data, "\n");
-	conn->errorMessage.len = strlen(conn->errorMessage.data);
-	if (tmpsock >= 0)
+	/*
+	 * Make sure we don't overflow the error buffer. Leave space for
+	 * the \n at the end, and for the terminating zero.
+	 */
+	maxlen = errbufsize - strlen(errbuf) - 2;
+	if (maxlen >= 0)
 	{
-		closesocket(tmpsock);
-#ifdef WIN32
-		WSASetLastError(save_errno);
-#else
-		errno = save_errno;
-#endif
+		strncat(errbuf, SOCK_STRERROR(SOCK_ERRNO, sebuf, sizeof(sebuf)),
+				maxlen);
+		strcat(errbuf, "\n");
 	}
+	if (tmpsock >= 0)
+		closesocket(tmpsock);
+	SOCK_ERRNO_SET(save_errno);
 	return FALSE;
 }
 
+/*
+ * PQcancel: request query cancel
+ *
+ * Returns TRUE if able to send the cancel request, FALSE if not.
+ *
+ * On failure, an error message is stored in *errbuf, which must be of size
+ * errbufsize (recommended size is 256 bytes).  *errbuf is not changed on
+ * success return.
+ */
+int
+PQcancel(PGcancel *cancel, char *errbuf, int errbufsize)
+{
+	if (!cancel)
+	{
+		StrNCpy(errbuf, "PQcancel() -- no cancel object supplied", errbufsize);
+		return FALSE;
+	}
+
+	return internal_cancel(&cancel->raddr, cancel->be_pid, cancel->be_key,
+						   errbuf, errbufsize);
+}
+
+/*
+ * PQrequestCancel: old, not thread-safe function for requesting query cancel
+ *
+ * Returns TRUE if able to send the cancel request, FALSE if not.
+ *
+ * On failure, the error message is saved in conn->errorMessage; this means
+ * that this can't be used when there might be other active operations on
+ * the connection object.
+ *
+ * NOTE: error messages will be cut off at the current size of the
+ * error message buffer, since we dare not try to expand conn->errorMessage!
+ */
+int
+PQrequestCancel(PGconn *conn)
+{
+	int r;
+
+	/* Check we have an open connection */
+	if (!conn)
+		return FALSE;
+
+	if (conn->sock < 0)
+	{
+		StrNCpy(conn->errorMessage.data,
+				"PQrequestCancel() -- connection is not open\n",
+				conn->errorMessage.maxlen);
+		conn->errorMessage.len = strlen(conn->errorMessage.data);
+
+		return FALSE;
+	}
+
+	r = internal_cancel(&conn->raddr, conn->be_pid, conn->be_key,
+						conn->errorMessage.data, conn->errorMessage.maxlen);
+
+	if (!r)
+		conn->errorMessage.len = strlen(conn->errorMessage.data);
+
+	return r;
+}
+
 
 /*
  * pqPacketSend() -- convenience routine to send a message to server.
@@ -2381,17 +2467,13 @@ pqPacketSend(PGconn *conn, char pack_type,
 
 
 
-#ifndef SYSCONFDIR
-#error "You must compile this file with SYSCONFDIR defined."
-#endif
-
 #define MAXBUFSIZE 256
 
 static int
 parseServiceInfo(PQconninfoOption *options, PQExpBuffer errorMessage)
 {
 	char	   *service = conninfo_getval(options, "service");
-	char	   *serviceFile = SYSCONFDIR "/pg_service.conf";
+	char		serviceFile[MAXPGPATH];
 	bool		group_found = false;
 	int			linenr = 0,
 				i;
@@ -2404,6 +2486,13 @@ parseServiceInfo(PQconninfoOption *options, PQExpBuffer errorMessage)
 	if (service == NULL)
 		service = getenv("PGSERVICE");
 
+	/*
+	 * This could be used by any application so we can't use the binary
+	 * location to find our config files.
+	 */
+	snprintf(serviceFile, MAXPGPATH, "%s/pg_service.conf",
+		   getenv("PGSYSCONFDIR") ? getenv("PGSYSCONFDIR") : SYSCONFDIR);
+
 	if (service != NULL)
 	{
 		FILE	   *f;
@@ -2790,7 +2879,7 @@ char *
 PQdb(const PGconn *conn)
 {
 	if (!conn)
-		return (char *) NULL;
+		return NULL;
 	return conn->dbName;
 }
 
@@ -2798,7 +2887,7 @@ char *
 PQuser(const PGconn *conn)
 {
 	if (!conn)
-		return (char *) NULL;
+		return NULL;
 	return conn->pguser;
 }
 
@@ -2806,7 +2895,7 @@ char *
 PQpass(const PGconn *conn)
 {
 	if (!conn)
-		return (char *) NULL;
+		return NULL;
 	return conn->pgpass;
 }
 
@@ -2814,7 +2903,7 @@ char *
 PQhost(const PGconn *conn)
 {
 	if (!conn)
-		return (char *) NULL;
+		return NULL;
 	return conn->pghost ? conn->pghost : conn->pgunixsocket;
 }
 
@@ -2822,7 +2911,7 @@ char *
 PQport(const PGconn *conn)
 {
 	if (!conn)
-		return (char *) NULL;
+		return NULL;
 	return conn->pgport;
 }
 
@@ -2830,7 +2919,7 @@ char *
 PQtty(const PGconn *conn)
 {
 	if (!conn)
-		return (char *) NULL;
+		return NULL;
 	return conn->pgtty;
 }
 
@@ -2838,7 +2927,7 @@ char *
 PQoptions(const PGconn *conn)
 {
 	if (!conn)
-		return (char *) NULL;
+		return NULL;
 	return conn->pgoptions;
 }
 
@@ -2885,6 +2974,16 @@ PQprotocolVersion(const PGconn *conn)
 	return PG_PROTOCOL_MAJOR(conn->pversion);
 }
 
+int
+PQserverVersion(const PGconn *conn)
+{
+	if (!conn)
+		return 0;
+	if (conn->status == CONNECTION_BAD)
+		return 0;
+	return conn->sversion;
+}
+
 char *
 PQerrorMessage(const PGconn *conn)
 {
@@ -2922,7 +3021,7 @@ int
 PQsetClientEncoding(PGconn *conn, const char *encoding)
 {
 	char		qbuf[128];
-	static char query[] = "set client_encoding to '%s'";
+	static const char query[] = "set client_encoding to '%s'";
 	PGresult   *res;
 	int			status;
 
@@ -2940,7 +3039,7 @@ PQsetClientEncoding(PGconn *conn, const char *encoding)
 	sprintf(qbuf, query, encoding);
 	res = PQexec(conn, qbuf);
 
-	if (res == (PGresult *) NULL)
+	if (res == NULL)
 		return -1;
 	if (res->resultStatus != PGRES_COMMAND_OK)
 		status = -1;
@@ -3095,8 +3194,8 @@ static char *
 PasswordFromFile(char *hostname, char *port, char *dbname, char *username)
 {
 	FILE	   *fp;
-	char	   *pgpassfile;
-	char	   *home;
+	char		homedir[MAXPGPATH];
+	char		pgpassfile[MAXPGPATH];
 	struct stat stat_buf;
 
 #define LINELEN NAMEDATALEN*5
@@ -3114,26 +3213,14 @@ PasswordFromFile(char *hostname, char *port, char *dbname, char *username)
 	if (port == NULL)
 		port = DEF_PGPORT_STR;
 
-	/* Look for it in the home dir */
-	home = getenv("HOME");
-	if (!home)
+	if (!pqGetHomeDirectory(homedir, sizeof(homedir)))
 		return NULL;
 
-	pgpassfile = malloc(strlen(home) + 1 + strlen(PGPASSFILE) + 1);
-	if (!pgpassfile)
-	{
-		fprintf(stderr, libpq_gettext("out of memory\n"));
-		return NULL;
-	}
-
-	sprintf(pgpassfile, "%s/%s", home, PGPASSFILE);
+	snprintf(pgpassfile, sizeof(pgpassfile), "%s/%s", homedir, PGPASSFILE);
 
 	/* If password file cannot be opened, ignore it. */
 	if (stat(pgpassfile, &stat_buf) == -1)
-	{
-		free(pgpassfile);
 		return NULL;
-	}
 
 #ifndef WIN32
 	/* If password file is insecure, alert the user and ignore it. */
@@ -3142,13 +3229,11 @@ PasswordFromFile(char *hostname, char *port, char *dbname, char *username)
 		fprintf(stderr,
 				libpq_gettext("WARNING: Password file %s has world or group read access; permission should be u=rw (0600)\n"),
 				pgpassfile);
-		free(pgpassfile);
 		return NULL;
 	}
 #endif
 
 	fp = fopen(pgpassfile, "r");
-	free(pgpassfile);
 	if (fp == NULL)
 		return NULL;
 
@@ -3183,3 +3268,82 @@ PasswordFromFile(char *hostname, char *port, char *dbname, char *username)
 
 #undef LINELEN
 }
+
+/*
+ * Obtain user's home directory, return in given buffer
+ *
+ * On Unix, this actually returns the user's home directory.  On Windows
+ * it returns the PostgreSQL-specific application data folder.
+ *
+ * This is essentially the same as get_home_path(), but we don't use that
+ * because we don't want to pull path.c into libpq (it pollutes application
+ * namespace)
+ */
+bool
+pqGetHomeDirectory(char *buf, int bufsize)
+{
+#ifndef WIN32
+	char		pwdbuf[BUFSIZ];
+	struct passwd pwdstr;
+	struct passwd *pwd = NULL;
+
+	if (pqGetpwuid(geteuid(), &pwdstr, pwdbuf, sizeof(pwdbuf), &pwd) != 0)
+		return false;
+	StrNCpy(buf, pwd->pw_dir, bufsize);
+	return true;
+
+#else
+	char		tmppath[MAX_PATH];
+
+	ZeroMemory(tmppath, sizeof(tmppath));
+	if (!SHGetSpecialFolderPath(NULL, tmppath, CSIDL_APPDATA, FALSE))
+		return false;
+	snprintf(buf, bufsize, "%s/postgresql", tmppath);
+	return true;
+#endif
+}
+
+/*
+ * To keep the API consistent, the locking stubs are always provided, even
+ * if they are not required.
+ */
+
+static void
+default_threadlock(int acquire)
+{
+#ifdef ENABLE_THREAD_SAFETY
+#ifndef WIN32
+	static pthread_mutex_t singlethread_lock = PTHREAD_MUTEX_INITIALIZER;
+
+#else
+	static pthread_mutex_t singlethread_lock = NULL;
+	static long mutex_initlock = 0;
+
+	if (singlethread_lock == NULL)
+	{
+		while (InterlockedExchange(&mutex_initlock, 1) == 1)
+			 /* loop, another thread own the lock */ ;
+		if (singlethread_lock == NULL)
+			pthread_mutex_init(&singlethread_lock, NULL);
+		InterlockedExchange(&mutex_initlock, 0);
+	}
+#endif
+	if (acquire)
+		pthread_mutex_lock(&singlethread_lock);
+	else
+		pthread_mutex_unlock(&singlethread_lock);
+#endif
+}
+
+pgthreadlock_t
+PQregisterThreadLock(pgthreadlock_t newhandler)
+{
+	pgthreadlock_t prev = pg_g_threadlock;
+
+	if (newhandler)
+		pg_g_threadlock = newhandler;
+	else
+		pg_g_threadlock = default_threadlock;
+
+	return prev;
+}
